/*
 * lattice.h
 *
 *  Created on: 07.11.2018
 *  Updated 12.2021
 *      Author: georg
 */

#ifndef LATTICE_H_
#define LATTICE_H_
#include <iostream>

/**
 * This class provides the basic lattice geometry.
 * It is written in a general way for arbitrary dimensions,
 * which makes it a little bit more complicated. For
 * simple usage it can be replaced by simple arrays as
 * lookup tables for the neighbors and coordinates.
 *
 */


typedef unsigned long Index;

template<int D>
class Lattice{
public:
	static const unsigned int dim=D;
	static const unsigned int numsitelinks=dim;
	typedef unsigned long Index;
	typedef Index Coord[dim];
	typedef Index Neighbours[2*D];
private:
	Neighbours* neighbourarray_;
	Coord sizes_;
	Index volume_;
public:
      Lattice(const Lattice&)=delete;
       const Lattice& operator=(const Lattice&)=delete;
       
	Lattice(const Coord& size):
		neighbourarray_(nullptr),sizes_(),volume_(0){
	  volume_=1;
      for(unsigned int i(0);i<dim;i++){
    	  volume_*=size[i];
    	  sizes_[i]=size[i];
      }
      neighbourarray_=new Neighbours[volume_];
      fillArray();
	}
	~Lattice(){
		delete[] neighbourarray_;
	}

	Index volume()const{return (volume_);}

	Index size(const Index i)const{return sizes_[i];}

	Index numberOfNeighbours()const{
		return (2*dim);
	}

	Index getNeighbor(const Index& ind,const unsigned int k)const{
		return (neighbourarray_[ind][k]);
	}


	Index up(const Index& ind,const unsigned int direction)const{
		return (neighbourarray_[ind][2*direction]);
	}

	Index down(const Index& ind,const unsigned int direction)const{
		return (neighbourarray_[ind][2*direction+1]);
	}

	Index arrayIndex(const Coord & coordinates)const{
		Index ret(0);
		for(unsigned int i(0);i<dim;i++){
			ret=ret*sizes_[i]+coordinates[i];
		}
		return (ret);
	}

	void arrayCoordinate(const Index & ind,Coord& ret)const{
		Index fakt(volume_);
		Index curind(ind);
		for(unsigned int i(0);i<dim;i++){
			fakt/=sizes_[i];
			ret[i]=curind/fakt;
			curind-=ret[i]*fakt;
		}
	}

	void printCoordinate(const Coord& in)const{
		if(dim==0){
			std::cout<<"()";
			return;
		}
		std::cout<<"(";
		for(unsigned int i(0);i+1<dim;i++){
			std::cout<<in[i]<<",";
		}
		std::cout<<in[dim-1]<<")";
	}
	void startCoordinate(Coord& loopindex)const{
		for(unsigned int i(dim);i--;){
			loopindex[i]=0;
		}
	}

	bool nextCoordinate(Coord& loopindex)const{
       return (nextCoordinate(loopindex,0));
	}
	bool nextCoordinate(Coord& loopindex,const unsigned int level)const{
		if(level+1==D){
			loopindex[level]++;
			return (loopindex[level]<sizes_[level]);
		}
		if(nextCoordinate(loopindex,level+1)){
			return (true);
		}
		loopindex[level]++;
		loopindex[level+1]=0;
		return (loopindex[level]<sizes_[level]);
	}
    void copyCoordinate(const Coord& in, Coord& out){
    	for(unsigned int i(dim);i--;){
    		out[i]=in[i];
    	}
    }
private:
	void fillArray(){
		Coord coord;
		Coord coord2;
		startCoordinate(coord);
	    bool run(true);
	    while(run){
	    	const Index siteindex(arrayIndex(coord));
	    	for(unsigned int i(dim);i--;){
	    		copyCoordinate(coord,coord2);
	    		coord2[i]=(coord2[i]+1)%sizes_[i];
	    		neighbourarray_[siteindex][2*i]=arrayIndex(coord2);
	    		copyCoordinate(coord,coord2);
	    		coord2[i]=(coord2[i]+sizes_[i]-1)%sizes_[i];
	    		neighbourarray_[siteindex][2*i+1]=arrayIndex(coord2);

	    	}
	    	run=nextCoordinate(coord);
	    }
	}

};



#endif /* LATTICE_H_ */
